/*
 OTFBrutus.cpp

 Copyright (c) 2010 Josh Harris (tateu)

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation and/or
     other materials provided with the distribution.
  3. The name of the author may not be used to endorse or promote products derived
     from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	2010-12-11 - JHH
		* v0.1.0.0 - Original release
*/

#include "OTFBrutus.h"


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
INT64 filesize(HANDLE hd) {
  BY_HANDLE_FILE_INFORMATION bhf;
  GetFileInformationByHandle(hd, &bhf);
  INT64 li = bhf.nFileSizeLow;
  if (bhf.nFileSizeHigh)
    li += ((INT64)bhf.nFileSizeHigh) << 32;
  return li;
}

char *timeFormatted(double sec, char *ret, int width = 8)
{
	//char ret[34];

	if (width < 8) {
		width = 8;
	} else if (width > 32) {
		width = 32;
	}

	int hrs = int(sec / 60 / 60);
	int mns = int(sec / 60) % 60;
	int scs = int(sec) % 60;

	//sprintf(ret, "%*s%02d:%02d:%02d", width - 8, " ", hrs, mns, scs);
	sprintf(ret, "%*d:%02d:%02d", width - 6, hrs, mns, scs);
	//sprintf(ret, "%02d:%02d:%02d", hrs, mns, scs);
	ret[width] = 0;
	return ret;
}

DWORD canAccessFile(char *fileName, bool bReadOnly, TCHAR *szErr)
{
	int openFlags = GENERIC_READ;
	int shareFlags = FILE_SHARE_READ;
	DWORD dwSize = 0, dwHigh = 0;
	if (!bReadOnly) {
		openFlags |= GENERIC_WRITE;
		shareFlags |= FILE_SHARE_WRITE;
	}

	HANDLE hFile = CreateFile(
					   fileName,
					   openFlags,
					   shareFlags,
					   NULL,
					   OPEN_EXISTING,
					   0,
					   NULL);

	if (hFile != INVALID_HANDLE_VALUE) {
		dwSize = GetFileSize(hFile, &dwHigh);
	} else {
		//DWORD dwError = GetLastError();

		/*if (!szErr) {
			szErr = new TCHAR[128];
		}

		if (dwError == ERROR_FILE_NOT_FOUND) {
			_stprintf_s(szErr, 128, _T("File Error: File Not Found"));
		}
		else if (dwError == ERROR_PATH_NOT_FOUND) {
			_stprintf_s(szErr, 128, _T("File Error: Path Not Found"));
		}
		else if (dwError == ERROR_ACCESS_DENIED) {
			_stprintf_s(szErr, 128, _T("File Error: Access Denied"));
		}
		else {
			_stprintf_s(szErr, 128, _T("File Error: Unkown Error"));
		}*/
	}

	CloseHandle(hFile);
	return dwSize;
}

// Threads
void passwordThread(void *arg)
{
	int i;
	for (pwd_idx = 0; pwd_idx < idxEnd && !bSuccess; pwd_idx++) {
		if (idxStart <= pwd_idx) {
			pwd.clear();
		}
		for (i = 0; i < m_vpwdParts.size(); i++)
		{
			int j;
			for (j = 0; j < m_vpwdParts[i].minLength; j++)
			{
				int curPos = m_vpwdParts[i].curPos;
				if (idxStart <= pwd_idx) {
					//pwd.push_back(m_vpwdParts[i].charSet[curPos].charItem.c_str());
					pwd += m_vpwdParts[i].charSet[curPos].charItem;
				}

				curPos = 0;
			}
		}

		if (idxStart <= pwd_idx) {
			int id = WAIT_OBJECT_0 + dwEvent;
			if (id >= 0 && id < threads) {
				memcpy(thread_params[id].pwd.Text, pwd.c_str(), pwd.size());
				thread_params[id].pwd.Length = pwd.size();
				thread_params[id].bRunning = TRUE;
				thread_params[id].bBoot = iVolType == OTF_VOL_BOOT ? TRUE : FALSE;
				ResetEvent(hEvent[id]);
				SetEvent(hEventThread[id]);
			} else {
				printf("\ndwEvent = %d\n", dwEvent);
			}

			dwEvent = WaitForMultipleObjects(
				threads, // number of objects in array
				hEvent,  // array of objects
				FALSE,   // wait for all
				5000);   // INFINITE

			counter++;
		}

		if (bComplete) {
			bSuccess = bComplete;
			break;
		}

		// increment the chars from last position first
		for (i = m_vpwdParts.size() - 1; i >= 0; i--)
		{
			 m_vpwdParts[i].curPos++;

			 if (m_vpwdParts[i].curPos >= m_vpwdParts[i].charSet.size()) {
				m_vpwdParts[i].curPos = 0;
			} else {
				break;
			}
		}
	}

	SetEvent(hEvent[threads]);
	SetEvent(hEventThread[threads]);
}

void dictionaryThread(void *arg)
{
	ifstream stream;
	stream.open(dictionary);
	if (stream.is_open()) {
		while (getline(stream, pwd)) {
			pwd_idx_dict++;
			//bSuccess = OTFE_Decrypt_Header(ciphertext, plaintext, 512, (byte *)pwd.c_str(), pwd.size());

			if (idxStart <= pwd_idx_dict) {
				int id = WAIT_OBJECT_0 + dwEvent;
				if (id >= 0 && id < threads) {
					memcpy(thread_params[id].pwd.Text, pwd.c_str(), pwd.size());
					thread_params[id].pwd.Length = pwd.size();
					thread_params[id].bRunning = TRUE;
					thread_params[id].bBoot = iVolType == OTF_VOL_BOOT ? TRUE : FALSE;
					ResetEvent(hEvent[id]);
					SetEvent(hEventThread[id]);
				} else {
					printf("\ndwEvent = %d\n", dwEvent);
				}

				dwEvent = WaitForMultipleObjects(
					threads, // number of objects in array
					hEvent,  // array of objects
					FALSE,   // wait for all
					5000);   // INFINITE

				counter++;
			}

			if (bComplete) {
				bSuccess = bComplete;
				break;
			}
		}
	} else {
		printf("ERROR Can't open dictionary:\n\t%s\n", dictionary);
		exit(1);
	}

	stream.close();

	SetEvent(hEvent[threads]);
	SetEvent(hEventThread[threads]);
}

void encryptionThread(void *arg)
{
	int nStatus = FALSE;
	DWORD dwWaitResult;
	PCRYPTO_INFO cryptoInfo = NULL;
	int id = (INT_PTR)arg;
	string temp_pwd;

start:
	dwWaitResult = WaitForSingleObject(hEventThread[id], INFINITE);
	if (dwWaitResult == WAIT_OBJECT_0)
	{
		if (!bComplete) {
			if (thread_params[id].bRunning) {
				ResetEvent(hEventThread[id]);

				thread_params[id].pwd.Text[thread_params[id].pwd.Length] = 0;
				temp_pwd = reinterpret_cast<const char*>(thread_params[id].pwd.Text);
				if (firstKeyFile) {
					KeyFilesApply (&thread_params[id].pwd, firstKeyFile);
				}

				if (bVerbose) {
					printf("Password: %s\n", thread_params[id].pwd.Text);
				}

				nStatus = ReadVolumeHeader (iVolType == OTF_VOL_BOOT ? TRUE : FALSE, headerCipher, &thread_params[id].pwd, &cryptoInfo, NULL, &cipher_set); //&cryptoInfo, cryptoInfo
				if (nStatus == ERR_CIPHER_INIT_WEAK_KEY) {
					nStatus = 0;	// We can ignore this error here
				}

				actual_pwds_tried++;

				/*if (nStatus == ERR_PASSWORD_WRONG) {
				} else if (nStatus != 0) {
				}*/
				if (nStatus == 0) {
					found_pwd = temp_pwd;
					bComplete = TRUE;
					ctx_hash_idx = cryptoInfo->pkcs5 - FIRST_PRF_ID;
					ctx_mode_idx = cryptoInfo->mode - FIRST_MODE_OF_OPERATION_ID;
					ctx_cipher_idx = cryptoInfo->ea - EAGetFirst ();
					for (int i = 0; i <= threads; i++) {
						SetEvent(hEvent[i]);
						SetEvent(hEventThread[i]);
					}
					_endthread();
				}

				SetEvent(hEvent[id]);
			}
		} else {
//			ResetEvent(hEvent[id]);
			ResetEvent(hEventThread[id]);
			_endthread();
		}
	}

	goto start;

	_endthread();
}
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
int main(int argc, char *argv[])
{
	HANDLE m_hVolume;
	DWORD dwVolAccess = GENERIC_READ;
	DWORD dwBytesRead = 0;
	ULARGE_INTEGER m_i64FileSz;
	LARGE_INTEGER m_iPos;
	byte inbuf[512];
	char durationStr[13];
	char timeLeftStr[13];

	LARGE_INTEGER nFreq, nStart, nEnd;

	char *container = NULL;

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

	printf("OTFBrutus v0.1.0.0\n");
	if (argc < 2) {
		printf("Please See OTFBrutus.txt, OTFBrutus_Options.txt and OTFBrutus_Sample.bat for useage:\n\n");
		exit(1);
	}

	int i;
	int j;
	for (i = 1; i + 1 < argc; i++) {
		if (argv[i][0] == '-' && strlen(argv[i]) == 2) {
			char t = argv[i][1];
			t = 'a';
			switch (argv[i][1]) {
				case 'i':
					container = argv[i+1];
					break;
				case 'p':
					pwd_pattern = argv[i+1];
					break;
				case 'd':
					dictionary = argv[i+1];
					break;
				case 'v':
					iVolType = atoi(argv[i+1]);
					break;
				case 'c':
					if (atoi(argv[i+1]) == OTF_CUR_CIPHERS) {
						for (j = 0; j <= OTF_TWOFISH_SERPENT; j++) {
							user_cipher.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_DEP_CIPHERS) {
						for (j = OTF_BLOWFISH; j <= OTF_IDEA; j++) {
							user_cipher.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_CIPHERS_V43) {
						for (j = 0; j <= OTF_TWOFISH_SERPENT; j++) {
							user_cipher.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_CIPHERS_V30) {
						for (j = 0; j <= OTF_AES_BLOWFISH_SERPENT; j++) {
							user_cipher.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_CIPHERS_V21) {
						for (j = 0; j <= OTF_DES3; j++) {
							if (j >= OTF_SERPENT && j <= OTF_TWOFISH_SERPENT) {
								continue;
							}
							user_cipher.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_CIPHERS_V20) {
						for (j = 0; j <= OTF_DES3; j++) {
							if (j >= OTF_SERPENT && j <= OTF_TWOFISH_SERPENT) {
								continue;
							}
							user_cipher.push_back(j);
						}
						user_cipher.push_back(OTF_IDEA);
					} else if (atoi(argv[i+1]) == OTF_CIPHERS_V10) {
						for (j = OTF_BLOWFISH; j <= OTF_DES3; j++) {
							user_cipher.push_back(j);
						}
						user_cipher.push_back(OTF_IDEA);
					} else {
						user_cipher.push_back(atoi(argv[i+1]));
					}
					break;
				case 'h':
					if (atoi(argv[i+1]) == OTF_CUR_HASHES) {
						for (j = 0; j <= OTF_WHIRLPOOL; j++) {
							user_hash.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_DEP_HASHES) {
						for (j = OTF_SHA1; j <= OTF_SHA1; j++) {
							user_hash.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_HASHES_V50) {
						for (j = 0; j <= OTF_WHIRLPOOL; j++) {
							user_hash.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_HASHES_V40) {
						for (j = 0; j <= OTF_SHA1; j++) {
							if (j == OTF_SHA512) {
								continue;
							}
							user_hash.push_back(j);
						}
					} else if (atoi(argv[i+1]) == OTF_HASHES_V21) {
						user_hash.push_back(OTF_RIPEMD160);
						user_hash.push_back(OTF_SHA1);
					} else if (atoi(argv[i+1]) == OTF_HASHES_V10) {
						user_hash.push_back(OTF_SHA1);
					} else {
						user_hash.push_back(atoi(argv[i+1]));
					}
					break;
				case 'm':
					if (atoi(argv[i+1]) == OTF_CUR_MODES) {
						user_mode.push_back(OTF_XTS);
					} else if (atoi(argv[i+1]) == OTF_DEP_MODES) {
						user_mode.push_back(OTF_LRW);
						user_mode.push_back(OTF_CBC);
					} else if (atoi(argv[i+1]) == OTF_MODES_V50) {
						user_mode.push_back(OTF_XTS);
					} else if (atoi(argv[i+1]) == OTF_MODES_V41) {
						user_mode.push_back(OTF_LRW);
					} else if (atoi(argv[i+1]) == OTF_MODES_V10) {
						user_mode.push_back(OTF_CBC);
					} else {
						user_mode.push_back(atoi(argv[i+1]));
					}
					break;
				case 'k':
					keyFiles.push_back(argv[i+1]);
					break;
				case 't':
					threads = atoi(argv[i+1]);
					break;
				case 's':
					idxStart = _atoi64(argv[i+1]);
					if (idxStart < 1)
						idxStart = 1;
					break;
				case 'e':
					idxEnd = _atoi64(argv[i+1]);
					if (idxEnd < 0)
						idxEnd = 0;
					break;
				case 'w':
					bVerbose = atoi(argv[i+1]) != 0;
					break;
				default:
					break;
			}
			i++;
		}
	}

	if (container == NULL || (pwd_pattern == NULL && dictionary == NULL)) {
		printf("Please See OTFBrutus.txt, OTFBrutus_Options.txt and OTFBrutus_Sample.bat for useage\n\n");
		exit(1);
	}

	if (pwd_pattern && dictionary) {
		printf("Cannot use pwd_pattern (-p) and dictionary (-d)\n\nPlease See OTFBrutus.txt, OTFBrutus_Options.txt and OTFBrutus_Sample.bat for useage\n\n");
		exit(1);
	}

	if (iVolType > OTF_VOL_HIDDEN_LEGACY1) {
		printf("Volume type (-v) cannot be greater than %d\n\nPlease See OTFBrutus.txt and OTFBrutus_Sample.bat for useage\n\n", OTF_VOL_HIDDEN_LEGACY1);
		exit(1);
	}

	// Make sure threads is between 1 and 32
	if (threads < 1)
		threads = 1;
	else if (threads > 32)
		threads = 32;

///////////////////////////////////////
// \\?\GLOBALROOT\Device\Harddisk0\Partition1
	m_hVolume = CreateFile(container, dwVolAccess,
				FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
				OPEN_EXISTING, FILE_FLAG_NO_BUFFERING,
				NULL);

	if (m_hVolume == INVALID_HANDLE_VALUE) {
		LPVOID lpMsgBuf;

        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      GetLastError(),
                      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                      (LPTSTR) &lpMsgBuf,
                      0,
                      NULL);

		printf("Error loading container\n%s\n%s\n", (LPCTSTR)lpMsgBuf, container);

		LocalFree(lpMsgBuf);
		//SecureZeroMemory(szPwd, sizeof(szPwd));
		return 3;
	}

	m_i64FileSz.LowPart = GetFileSize(m_hVolume, &m_i64FileSz.HighPart);

	if (iVolType == OTF_VOL_BOOT)
	{
		user_hash.clear();
		user_hash.push_back(OTF_RIPEMD160); //The only hash allowed for system encryption as of TC 7.0a
		user_mode.clear();
		user_mode.push_back(OTF_XTS); //The only mode allowed for system encryption as of TC 7.0a

		if (m_i64FileSz.QuadPart < BOOT_ISO_OFFSET) {
			m_iPos.QuadPart = 0; //Assume 512 byte Header only
		} else {
			if (m_i64FileSz.QuadPart == ISO_SIZE) {
				m_iPos.QuadPart = BOOT_ISO_OFFSET; //Rescue Disc
			} else {
				m_iPos.QuadPart = 62 * TC_SECTOR_SIZE_BIOS; //Device
			}

			if (!SetFilePointer(m_hVolume, m_iPos.LowPart, &m_iPos.HighPart, FILE_BEGIN)) {
				printf("Cannot seek to boot volume header position\n");
				CloseHandle(m_hVolume);
				return 3;
			}
		}
	}
	else if (iVolType)
	{
		if (m_i64FileSz.QuadPart <= BACKUP_H_SIZE_LEGACY)
			m_iPos.QuadPart = -TC_VOLUME_HEADER_SIZE_LEGACY; //Old TC Header format, backup header file is 1024 bytes
		else if (m_i64FileSz.QuadPart == BACKUP_H_SIZE)
			m_iPos.QuadPart = TC_HIDDEN_VOLUME_HEADER_OFFSET; //New Header Format, backup header file is 131072 bytes
		else if (iVolType == OTF_VOL_HIDDEN_LEGACY1)
			m_iPos.QuadPart = -TC_HIDDEN_VOLUME_HEADER_OFFSET_LEGACY; //File or partition, Legacy, TrueCrypt version prior to 6.0
		else
			m_iPos.QuadPart = TC_HIDDEN_VOLUME_HEADER_OFFSET; //File or partition

		int seekStart = m_iPos.QuadPart > 0 ? FILE_BEGIN : FILE_END;
		if (!SetFilePointer(m_hVolume, m_iPos.LowPart, &m_iPos.HighPart, seekStart)) {
			printf("Cannot seek to hidden volume header position\n");
			CloseHandle(m_hVolume);
			//SecureZeroMemory(szPwd, sizeof(szPwd));
			return 3;
		}
	}

	if (!ReadFile(m_hVolume, inbuf, TC_SECTOR_SIZE_LEGACY, &dwBytesRead, FALSE)) {
		printf("Cannot read volume header\n");
		CloseHandle(m_hVolume);
		return 3;
	}

	memcpy(headerCipher, inbuf, TC_SECTOR_SIZE_LEGACY);
	CloseHandle(m_hVolume);

	/////////////////////////////////
	//
	if (pwd_pattern == NULL)
	{
		pwd_pattern_len = 0;
		pwd_count = 0;
	}
	else
	{
		pwd_pattern_len = strlen(pwd_pattern);

		for (i = 0; i < pwd_pattern_len - 1; i++)
		{
			if (pwd_pattern[i] == '[')
			{
				int pattern_count = 1;
				pwdParts.charSet.clear();
				charSet.charItem.clear();

				// increase pos counter to move past the opening '['
				i++;
				while (pwd_pattern[i] != ']' && i < pwd_pattern_len - 1)
				{
					if (pwd_pattern[i] == '\\')
						i++;

					charSet.charItem.clear();
					charSet.charItem.push_back(pwd_pattern[i]);
					charSet.pos = 0;
					pwdParts.charSet.push_back(charSet);

					i++;
				}

				if (pwd_pattern[i] != ']') {
					printf("ERROR at pos %d: did not find closing ']'\n", i);
					return 10;
				}

				//Check for length character bracket open '{'
				if (i < pwd_pattern_len - 3) {
					int k = 1;
					string pCountStr;
					if (pwd_pattern[i + k] == '{') {
						k++;
						while (pwd_pattern[i + k] != '}' && i + k < pwd_pattern_len - 1) {
							pCountStr.push_back(pwd_pattern[i + k]);
							k++;
						}

						if (pwd_pattern[i + k] == '}') {
							pattern_count = atoi(pCountStr.c_str());
							i += k;
						} else {
							printf("ERROR at pos %d: did not find closing '}'\n", i + k);
							return 10;
						}
					}
				}

				pwdParts.curPos = 0;
				pwdParts.minLength = 1;
				pwdParts.maxLength = 1;

				if (pwdParts.charSet.size() > 0) {
					for (int k = 0; k < pattern_count; k++) {
						m_vpwdParts.push_back(pwdParts);
					}
				}

				pwdParts.charSet.clear();
			}// END []
			else if (pwd_pattern[i] == '(')
			{
				int pattern_count = 1;
				pwdParts.charSet.clear();
				charSet.charItem.clear();

				// increase pos counter to move past the opening '('
				i++;
				while (pwd_pattern[i] != ')' && i < pwd_pattern_len - 1)
				{
					if (pwd_pattern[i] == '\\')
						i++;
					else if (pwd_pattern[i] == '|')
					{
						i++;
						charSet.pos = 0;

						if (charSet.charItem.size() > 0) {
							pwdParts.charSet.push_back(charSet);
						}

						charSet.charItem.clear();
						continue;
					}


					charSet.charItem.push_back(pwd_pattern[i]);

					i++;
				}

				if (pwd_pattern[i] != ')') {
					printf("ERROR at pos %d: did not find closing ')'\n", i);
					return 10;
				}

				//Check for length character bracket open '{'
				if (i < pwd_pattern_len - 3) {
					int k = 1;
					string pCountStr;
					if (pwd_pattern[i + k] == '{') {
						k++;
						while (pwd_pattern[i + k] != '}' && i + k < pwd_pattern_len - 1) {
							pCountStr.push_back(pwd_pattern[i + k]);
							k++;
						}

						if (pwd_pattern[i + k] == '}') {
							pattern_count = atoi(pCountStr.c_str());
							i += k;
						} else {
							printf("ERROR at pos %d: did not find closing '}'\n", i + k);
							return 10;
						}
					}
				}

				charSet.pos = 0;
				if (charSet.charItem.size() > 0) {
					pwdParts.charSet.push_back(charSet);
				}

				charSet.charItem.clear();

				pwdParts.curPos = 0;
				pwdParts.minLength = 1;
				pwdParts.maxLength = 1;

				if (pwdParts.charSet.size() > 0) {
					for (int k = 0; k < pattern_count; k++) {
						m_vpwdParts.push_back(pwdParts);
					}
				}

				pwdParts.charSet.clear();
			} // END ()
			else if (pwd_pattern[i] == ']')
			{
				printf("ERROR at pos %d: Found ']' that was not part of a pattern or escaped \\\n", i);
				return 10;
			}
			else if (pwd_pattern[i] == ')')
			{
				printf("ERROR at pos %d: Found ')' that was not part of a pattern or escaped \\\n", i);
				return 10;
			}
			else if (pwd_pattern[i] == '{')
			{
				printf("ERROR at pos %d: Found '{' that was not part of a pattern or escaped \\\n", i);
				return 10;
			}
			else if (pwd_pattern[i] == '}')
			{
				printf("ERROR at pos %d: Found '}' that was not part of a pattern or escaped \\\n", i);
				return 10;
			}
			else
			{
				pwdParts.charSet.clear();
				charSet.charItem.clear();

				while (i < pwd_pattern_len - 1)
				{
					if (pwd_pattern[i] == '\\')
					{
						i++;
						continue;
					}
					else if (pwd_pattern[i] == '[' || pwd_pattern[i] == '(' || pwd_pattern[i] == ']' || pwd_pattern[i] == ')' || pwd_pattern[i] == '{' || pwd_pattern[i] == '}')
					{
						i--; // decrease counter so the for loop finds the char
						break;
					}

					charSet.charItem.push_back(pwd_pattern[i]);

					i++;
				}

				charSet.pos = 0;
				if (charSet.charItem.size() > 0)
					pwdParts.charSet.push_back(charSet);

				charSet.charItem.clear();

				pwdParts.curPos = 0;
				pwdParts.minLength = 1;
				pwdParts.maxLength = 1;

				if (pwdParts.charSet.size() > 0)
					m_vpwdParts.push_back(pwdParts);

				pwdParts.charSet.clear();
			} // END any chars
		} // END for (i = 0; i < pwd_pattern_len - 1; i++)
	} // END if (pwd_pattern == NULL)

	if (m_vpwdParts.size() > 0) {
		pwd_count = 1;
		for (i = 0; i < m_vpwdParts.size(); i++)
		{
			pwd_count *= pow((double)m_vpwdParts[i].charSet.size(), (double)m_vpwdParts[i].minLength);
		}
	}
	//
	/////////////////////////////////

	/////////////////////////////////
	//
	printf("Attempting to read the");
	switch (iVolType) {
		case OTF_VOL_TYPE_STD:
			printf(" Standard");
			break;
		case OTF_VOL_HIDDEN:
			printf(" Hidden");
			break;
		case OTF_VOL_BOOT:
			printf(" System");
			break;
		case OTF_VOL_HIDDEN_LEGACY1:
			printf(" Hidden Legacy");
			break;
		default:
			printf(" Standard");
			break;
	}

	printf(" Volume from a");
	if (_strnicmp(container, "\\\\?", 3) == 0) {
		printf(" Device");
	} else {
		printf(" File");
	}

	printf(" hosted");
	if (iVolType == OTF_VOL_BOOT) {
		if (m_i64FileSz.QuadPart > ISO_SIZE) {
			printf(" Container");
		} else {
			printf(" Rescue Disc");
		}
	} else if (iVolType == OTF_VOL_HIDDEN_LEGACY1) {
		if (m_i64FileSz.QuadPart > BACKUP_H_SIZE_LEGACY) {
			printf(" Container");
		} else {
			printf(" Header");
		}
	} else {
		if (m_i64FileSz.QuadPart > BACKUP_H_SIZE) {
			printf(" Container");
		} else {
			printf(" Header");
		}
	}
	printf(":\n\t%s\n", container);

	if (dictionary != NULL) {
		if (!canAccessFile(dictionary, true, NULL)) {
			printf("ERROR cannot read dictionary\n\t%s\n", dictionary);
			return 9;
		}
		printf("Using a custom word list:\n\t%s\n", dictionary);
		printf("...Please wait, Counting dictionary list now...\r");

		FILE *f = fopen(dictionary, "rb");
		int b;
		while((b = fgetc(f))!=EOF)
			pwd_count_dict += (b==10) ? 1 : 0;
		fclose(f);

		printf("                                               \r");
		printf("containing:\n\t%I64d passwords\n", pwd_count_dict);
	}
	if (pwd_count > 0)
		printf("Using a password pattern:\n\t%s\ncontaining:\n\t%I64d passwords\n", pwd_pattern, pwd_count);

	if (pwd_count_dict > 0)
		pwd_count = pwd_count_dict;

	idxStart--;
	if (idxStart > 0 || idxEnd > 0) {
		if (idxEnd > pwd_count || idxEnd <= 0) {
			idxEnd = pwd_count;
		}
		if (idxStart > idxEnd) {
			idxStart = idxEnd - 1;
		}
		printf("  subset contains:\n\t%I64d passwords (%I64d - %I64d)\n",
			idxEnd - idxStart, idxStart + 1, idxEnd);
	}

	if (keyFiles.size() > 0)
		printf("Using key files:\n");
	for (i = 0; i < keyFiles.size(); i++) {
		if (!canAccessFile((char *)keyFiles[i].c_str(), true, NULL)) {
			printf("ERROR cannot read keyfile\n\t%s\n", keyFiles[i].c_str());
			return 9;
		}
		printf("\t%s\n", keyFiles[i].c_str());
		KeyFile *kf = (KeyFile *)malloc(sizeof(KeyFile));
		memcpy(kf->FileName, keyFiles[i].c_str(), sizeof(kf->FileName));
		firstKeyFile = KeyFileAdd(firstKeyFile, kf);
	}

	if (idxEnd == 0) {
		idxEnd = pwd_count;
	}

	cipher_set.cipher_count = 0;
	cipher_set.hash_count = 0;
	cipher_set.mode_count = 0;

	printf("Cipher:\n");
	if (user_cipher.size() == 0)
	{
		for (i = 0; i < TC_CIPHER_COUNT + TC_MULTICIPHER_COUNT; i++) {
			printf("\t%s\n", cipherNames[i]);
		}
	}
	else
	{
		sort(user_cipher.begin(), user_cipher.end());
		if (user_cipher[0] == OTF_ALL_CIPHERS)
		{
			user_cipher.clear();
			for (i = 0; i < TC_CIPHER_COUNT + TC_MULTICIPHER_COUNT; i++)
				printf("\t%s\n", cipherNames[i]);
		} else {
			user_cipher.erase(unique(user_cipher.begin(), user_cipher.end()), user_cipher.end()); //Delete dupes

			for (i = 0; i < user_cipher.size(); i++) {
				if ((user_cipher[i] >= TC_CIPHER_COUNT + TC_MULTICIPHER_COUNT) || user_cipher[i] < 0) {
					printf("\tWARNING: Skipping illegal cipher value %d\n", user_cipher[i]);
					user_cipher.erase(user_cipher.begin() + i);
					i--;
					continue;
				}
				printf("\t%s\n", cipherNames[user_cipher[i]]);
			}
		}

		cipher_set.cipher_count = user_cipher.size();
		for (i = 0; i < user_cipher.size(); i++) {
			cipher_set.cipher[i] = user_cipher[i];
		}
	}

	printf("Hash:\n");
	if (user_hash.size() == 0)
	{
		for (i = 0; i < TC_HASH_COUNT; i++) {
			printf("\t%s\n", hashNames[i]);
		}
	}
	else
	{
		sort(user_hash.begin(), user_hash.end());
		if (user_hash[0] == OTF_ALL_HASHES)
		{
			user_hash.clear();
			for (i = 0; i < TC_HASH_COUNT; i++)
				printf("\t%s\n", hashNames[i]);
		} else {
			user_hash.erase(unique(user_hash.begin(), user_hash.end()), user_hash.end()); //Delete dupes

			for (i = 0; i < user_hash.size(); i++) {
				if ((user_hash[i] >= TC_HASH_COUNT) || user_hash[i] < 0) {
					printf("\tWARNING: Skipping illegal hash value %d\n", user_hash[i]);
					user_hash.erase(user_hash.begin() + i);
					i--;
					continue;
				}
				printf("\t%s\n", hashNames[user_hash[i]]);
			}
		}

		cipher_set.hash_count = user_hash.size();
		for (i = 0; i < user_hash.size(); i++) {
			cipher_set.hash[i] = user_hash[i];
		}
	}

	printf("Mode:\n");
	if (user_mode.size() == 0)
	{
		for (i = 0; i < TC_MODE_COUNT; i++) {
			printf("\t%s\n", modeNames[i]);
		}
	}
	else
	{
		sort(user_mode.begin(), user_mode.end());
		if (user_mode[0] == OTF_ALL_MODES)
		{
			user_mode.clear();
			for (i = 0; i < TC_MODE_COUNT; i++)
				printf("\t%s\n", modeNames[i]);
		} else {
			user_mode.erase(unique(user_mode.begin(), user_mode.end()), user_mode.end()); //Delete dupes

			for (i = 0; i < user_mode.size(); i++) {
				if ((user_mode[i] >= TC_MODE_COUNT) || user_mode[i] < 0) {
					printf("\tWARNING: Skipping illegal mode value %d\n", user_mode[i]);
					user_mode.erase(user_mode.begin() + i);
					i--;
					continue;
				}
				printf("\t%s\n", modeNames[user_mode[i]]);
			}
		}

		cipher_set.mode_count = user_mode.size();
		for (i = 0; i < user_mode.size(); i++) {
			cipher_set.mode[i] = user_mode[i];
		}
	}
	//
	/////////////////////////////////

	/////////////////////////////////
	//
	for (int i = 0; i < threads; i++) {
		hEvent[i] = CreateEvent(NULL, TRUE, TRUE, NULL);
		if (hEvent[i] == INVALID_HANDLE_VALUE) {
			printf("Failed to create hEvent %d\n", i);
			exit(1);
		}
		hEventThread[i] = CreateEvent(NULL, TRUE, FALSE, NULL);
			if (hEventThread[i] == INVALID_HANDLE_VALUE) {
				printf("Failed to create hEventThread %d\n", i);
				exit(1);
			}
		hThread[i] = (HANDLE) _beginthread(encryptionThread, 0, (void*)i);
			if (hThread[i] == INVALID_HANDLE_VALUE) {
				printf("Failed to create hThread %d\n", i);
				exit(1);
			}

		thread_params[i].bRunning = FALSE;
	}

	printf("\n"
			"%7s  "
			"%12s  "
			"%12s  "
			"%10s  "
			"%9s\n",
			"Percent",
			"Duration",
			"TimeLeft",
			"Speed",
			"Count"
		);

	QueryPerformanceFrequency(&nFreq);
	QueryPerformanceCounter(&nStart);

	// Set this to TRUE, it is from crypto.cpp and will make sure that only 512 header block is tried at a time
	bOneHeader = TRUE;

	hEvent[threads] = CreateEvent(NULL, TRUE, TRUE, NULL);
	if (hEvent[threads] == INVALID_HANDLE_VALUE) {
		printf("Failed to create password hEvent %d\n", i);
		exit(1);
	}
	hEventThread[threads] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEventThread[threads] == INVALID_HANDLE_VALUE) {
			printf("Failed to create password hEventThread %d\n", i);
			exit(1);
		}

	if (dictionary) {
		pwd_count = 0; //Set to 0 so that pwd_pattern can't run

		hThread[threads] = (HANDLE) _beginthread(dictionaryThread, 0, (void*)0);
			if (hThread[threads] == INVALID_HANDLE_VALUE) {
				printf("Failed to create password hThread %d\n", i);
				exit(1);
			}

		DWORD dwWaitResult;
		while (WAIT_OBJECT_0 != WaitForSingleObject(hEventThread[threads], 1000)) {
			QueryPerformanceCounter(&nEnd);
			double duration = 1.0 * (nEnd.QuadPart - nStart.QuadPart) / nFreq.QuadPart;
			double pps = counter / duration;
			string ppsStr;
			//sprintf(&ppsStr, "%0.2f", idx / duration);
			double timeLeft = (idxEnd - pwd_idx_dict) / pps;
			double percentage = ((double)counter / (idxEnd - idxStart)) * 100.0;

			timeFormatted(duration, (char *)&durationStr, 12);
			timeFormatted(timeLeft, (char *)&timeLeftStr, 12);
			printf(
				"%6.0f%s  "
				"%12s  " //"%12.2f  "
				"%12s  "
				"%8.2f/s  "
				"%9I64d%s",
				percentage, "%",
				durationStr, //timeFormatted(duration)
				timeLeftStr,
				pps,
				counter, bVerbose ? "\n" : "\r"
			);
		}
	}
	//
	/////////////////////////////////

	/////////////////////////////////
	// Subtract the fake print increment count
	actual_pwds_tried -= print_count;
	print_count = 0;

	if (pwd_count > 0) {
		// Make sure events are in the signaled state
		for (int i = 0; i < threads; i++) {
			SetEvent(hEvent[i]);
		}

		hThread[threads] = (HANDLE) _beginthread(passwordThread, 0, (void*)0);
			if (hThread[threads] == INVALID_HANDLE_VALUE) {
				printf("Failed to create password hThread %d\n", threads);
				exit(1);
			}

		DWORD dwWaitResult;
		while (WAIT_OBJECT_0 != WaitForSingleObject(hEventThread[threads], 1000)) {
			QueryPerformanceCounter(&nEnd);
			double duration = 1.0 * (nEnd.QuadPart - nStart.QuadPart) / nFreq.QuadPart;
			double pps = counter / duration;
			string ppsStr;
			//sprintf(&ppsStr, "%0.2f", idx / duration);
			double timeLeft = (idxEnd - pwd_idx) / pps;
			double percentage = ((double)counter / (idxEnd - idxStart)) * 100.0;

			timeFormatted(duration, (char *)&durationStr, 12);
			timeFormatted(timeLeft, (char *)&timeLeftStr, 12);
			printf(
				"%6.0f%s  "
				"%12s  " //"%12.2f  "
				"%12s  "
				"%8.2f/s  "
				"%9I64d%s",
				percentage, "%",
				durationStr, //timeFormatted(duration)
				timeLeftStr,
				pps,
				counter, bVerbose ? "\n" : "\r"
			);
		}
	}
	//
	/////////////////////////////////

	/////////////////////////////////
	//
	QueryPerformanceCounter(&nEnd);

	dwEvent = WaitForMultipleObjects(
		threads,		// number of objects in array
		hEvent,			// array of objects
		TRUE,			// wait for all
		5000);
	// Subtract the fake print increment count
	actual_pwds_tried -= print_count;
	pwd_idx = actual_pwds_tried;

	double duration = 1.0 * (nEnd.QuadPart - nStart.QuadPart) / nFreq.QuadPart;
	double pps = counter / duration;
	string ppsStr;
	//sprintf(&ppsStr, "%0.2f", idx / duration);
	double timeLeft = 0; //(idxEnd - pwd_idx) / pps;
	double percentage = ((double)counter / (idxEnd - idxStart)) * 100.0;

	timeFormatted(duration, (char *)&durationStr, 12);
	timeFormatted(timeLeft, (char *)&timeLeftStr, 12);
	printf(
		"%6.0f%s  "
		"%12s  " //"%12.2f  "
		"%12s  "
		"%8.2f/s  "
		"%9I64d%s",
		percentage, "%",
		durationStr, //timeFormatted(duration)
		timeLeftStr,
		pps,
		counter, "\n"
	);

	if (bComplete) {
		printf("\nPassword Found\n");
		printf("\t%12s", "VolumeType:");
		switch (iVolType) {
			case OTF_VOL_TYPE_STD:
				printf(" Standard");
				break;
			case OTF_VOL_HIDDEN:
				printf(" Hidden");
				break;
			case OTF_VOL_BOOT:
				printf(" System");
				break;
			case OTF_VOL_HIDDEN_LEGACY1:
				printf(" Hidden Legacy");
				break;
			default:
				printf(" Standard");
				break;
		}
		printf("\n");

		printf("\t%12s %s\n", "password:", found_pwd.c_str());

		printf("\t%12s %s", "Cipher:", cipherNames[ctx_cipher_idx]);
		printf("\n");

		printf("\t%12s %s", "Hash:", hashNames[ctx_hash_idx]);
		printf("\n");

		printf("\t%12s %s", "Mode:", modeNames[ctx_mode_idx]);
		printf("\n");
	} else {
		printf("\nPassword Not Found\n");
	}
}
